"""
Definitions to support lazy expressions.

Copyright (c) 2024 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""
from typing import Callable, Optional, override, Any

import FuPy.tracing as tr
from FuPy import utils
from FuPy.basics import *
from FuPy.tracing_laziness import *

__all__ = [
    "Lazy", "IterLazy", "evaluate", "lazy", "lazyf",
]


class Lazy[A]:
    """A lazy value of type A, with memoization upon evaluation.

    See: https://en.wikipedia.org/wiki/Lazy_evaluation
    A.k.a. as thunk: https://en.wikipedia.org/wiki/Thunk
    Sharing of lazy values is encouraged, since they need to be evaluated only once.

    Usage:

    - Construct by v = Lazy(lambda: expression)

      N.B. the lambda captures local names occurring in expression, in its closure

    - Get value by evaluate(v); for internal use only: v._get()

    """
    next_thunk_number = 0
    # full_regime = False  # laziness regime: full or minimal

    def __init__(self, value: Callable[[], A] | A, name: Optional[str] = None) -> None:
        self.value = value  # unevaluated expression or cached evaluation result of the expression
        self.name = name
        self.nesting_count = 0  # number of times that nested Lazy expression is copied
        self.hit_count = 0  # number of times the cached value was used; 0 means unevaluated
        self.thunk_number = Lazy.next_thunk_number
        Lazy.next_thunk_number += 1
        tr.trace_step(lambda: SuspensionStep(f"{self}"))

    @override
    def __repr__(self) -> str:
        if self.nesting_count:
            nesting_state = f", nesting_count={self.nesting_count}"
        else:
            nesting_state = ""
        if self.hit_count:
            state = f"value={utils.show_value(self.value)}, thunk_number={self.thunk_number}{nesting_state}, hit_count={self.hit_count}"
        else:
            state = f"unevaluated={self.name or '`...`'}, thunk_number={self.thunk_number}"
        return f"Lazy({state})"

    def __str__(self) -> str:
        if self.hit_count:
            status = f"= {utils.show_value(self.value)}"
        else:
            status = f": {self.name or '`...`'}"
        return f"(θ_{self.thunk_number} {status})"

    def _get(self) -> A:
        """Evaluate lazy value, with caching/memoization.
        Repeats until result is not Lazy.
        This is useful to implement tail call elimination.
        """
        unevaluated = not self.hit_count
        nesting_comment = ""

        while unevaluated:
            tr.trace_step(lambda: EvaluationStep(f"{self}{nesting_comment}"))
            tr.inc_depth()
            self.value = self.value()  # take off lambda: wrapper; result could be Lazy again
            tr.dec_depth()
            if isinstance(self.value, Lazy):  # flatten, avoiding recursion (that would burden the stack)
                self.nesting_count += 1
                nesting_comment = f" {{nesting count: {self.nesting_count}}}" if self.nesting_count else ''
                unevaluated = not self.value.hit_count
                self.name = self.value.name
                self.value = self.value.value
            else:
                unevaluated = False

        self.hit_count += 1
        tr.trace_step(lambda: GettingStep(f"{self}"))
        return self.value

    def __call__[A, B](self: "Lazy[Callable[[A], B]]", arg) -> "Lazy[B]":
        """Apply as function.

        Assumption: self is Func instance

        For Weak Head Normal Form, the function itself is evaluated first.
        """
        # if Lazy.full_regime:
        #     return Lazy[B](lambda: lambda a: self._get()(a))
        # else:
        self._get()
        if not callable(self.value):
            raise TypeError(f"value in {self!r} is not callable")
        return self.value(arg)  # wrap this in Lazy again?

    def __add__(self, other):
        # if Lazy.full_regime:
        #     print("LAZY ADD")
        #     return Lazy(lambda: self._get() + other)
        # else:
        return self._get() + other

    def __radd__(self, other):
        # if Lazy.full_regime:
        #     return Lazy(lambda: other + self._get())
        # else:
        return other + self._get()

    def __sub__(self, other):
        # if Lazy.full_regime:
        #     return Lazy(lambda: self._get() - other)
        # else:
        return self._get() - other

    def __rsub__(self, other):
        # if Lazy.full_regime:
        #     return Lazy(lambda: other - self._get())
        # else:
        return other + self._get()

    def __mul__(self, other):
        # if Lazy.full_regime:
        #     return Lazy(lambda: self._get() * other)
        # else:
        return self._get() * other

    def __rmul__(self, other):
        # if Lazy.full_regime:
        #     return Lazy(lambda: other * self._get())
        # else:
        return other * self._get()

    def __getitem__(self, item):
        """Subscription triggers evaluation.
        Assumption: type A is subscriptable
        """
        # if Lazy.full_regime:
        #     return Lazy(lambda: self._get()[item])
        # else:
        return self._get()[item]

    # TODO: support other operators that automatically evaluate: __add__, etc.


type IterLazy[A] = A | Lazy[IterLazy[A]]  # for use in case of Tail Call Elimination


@func
def evaluate[A](v: Lazy[A] | A) -> A:
    """Evaluate v.
    """
    if isinstance(v, Lazy):
        return v._get()
    else:
        return v


def lazy[A, B](expr: Callable[[A], B] | str) -> Func[A, Lazy[B]] | Lazy[Any]:
    """Make expr lazy.

    Assumption: if expr is a string, then it doesn't contain local variables
    TODO: capture local variables; la needs to look one further stack frame up.

    If expr is callable, then expr(a) is equivalent to Lazy(lambda: expr(a)).
    """
    if callable(expr):
        expr = func(expr)
        return Func(lambda arg: Lazy(lambda: expr(arg), name=f"`{expr}({utils.show_value(arg)})`"))
    else:  # string
        return Lazy(la(f": {expr}"), name=f"`{expr}`")


def lazyf[X, Y](f: Callable[[X], Y]) -> Callable[[Lazy[X]], Lazy[Y]]:
    """Apply Lazy as functor to f.
    That is, apply f inside Lazy.
    """
    return Func(lambda x: Lazy(lambda: f(x._get())))



""" tracing_basics.py

Definitions to support tracing of basic functionality.

Copyright (c) 2024 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""
from typing import override
from FuPy import utils
import FuPy.tracing as tr

__all__ = [
    "DefinitionStep", "ApplicationStep", "MotivationStep", "ResultStep",
]


class DefinitionStep(tr.BaseStep):
    """Class for function construction steps.
    """
    def __init__(self, note: str) -> None:
        super().__init__(f"{note}")

    @override
    def __str__(self) -> str:
        return f"Define:    {super().__str__()}"


class ApplicationStep(tr.BaseStep):
    """Class for function application steps.
    """
    def __init__(self, note: str) -> None:
        super().__init__(utils.indent_lines(f"┌ {note}", '│ ', k=1))

    @override
    def __str__(self) -> str:
        return f"Apply:     {super().__str__()}"


class MotivationStep(tr.BaseStep):
    """Class for motivation steps.
    """
    def __init__(self, note: str) -> None:
        super().__init__(f"=   {{ {note} }}")

    @override
    def __str__(self) -> str:
        return f"Motivate:  {super().__str__()}"


class ResultStep(tr.BaseStep):
    """Class for result steps.
    """
    def __init__(self, note: str) -> None:
        super().__init__(f"└ {note}")

    @override
    def __str__(self) -> str:
        return f"Return:    {super().__str__()}"



"""
Utilities to support functional programming in Python.

Copyright (c) 2024 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""
import inspect
from typing import Any, Callable

__all__ = [
    "count_required_args",
    "indent_lines",
    "show_value", "show_args",
]


def count_required_args(f: Callable[..., Any]) -> int:
    """Count the minimum number of required non-keyword arguments of f.
    Does not work for built-in functions (use parameter r to bypass this).
    """
    from FuPy.basics import Func
    if isinstance(f, Func):
        return 1

    sig = inspect.signature(f)

    def parameter_is_required(p: inspect.Parameter) -> bool:
        return p.default is inspect.Parameter.empty and p.kind in (
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
        )

    return sum(parameter_is_required(p) for p
               in sig.parameters.values())


def indent_lines(line: str, indentation: str = '', k: int = 0) -> str:
    """Indent each line by n spaces, except for the first k lines.
    """
    lines = line.split('\n')
    return '\n'.join([f"{'' if index < k else indentation}{line}" for index, line in enumerate(lines)])


def show_value(value) -> str:
    """Show value, quoting strings.
    """
    # print(f"show_value(repr={value!r}, str={value!s})")
    if isinstance(value, tuple) and not value:
        return "_"
    elif isinstance(value, str):
        return repr(value)
    else:
        return str(value)


def show_args(args: tuple[Any, ...]) -> str:
    """Show tuple as a string, with strings quoted, without trailing comma if singleton.

    N.B. The singleton case does not occur in case of auto-(un)packing of arguments.
    """
    if not args:
        return "_"
    elif len(args) == 1:
        from FuPy.basics import Func
        arg = args[0]
        if isinstance(arg, Func):
            return arg.str('.')
        else:
            return show_value(arg)
    else:
        return f"({', '.join(show_value(arg) for arg in args)})"



"""
Definitions to support various fixpoints, both on the type level and the value level.

Copyright (c) 2024 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""
from dataclasses import dataclass
from typing import Final, Callable, Protocol

from FuPy.basics import *
from FuPy.laziness import *

__all__ = [
    "fix",
    "Functor", "fmap", "Fmap",
    "Fix", "in_", "out",
    "F_Algebra", "cata", "cata_",
    "F_CoAlgebra", "ana", "ana_",
    "hylo"
]


@func
def fix[A](f: Callable[[A], A]) -> A:
    """Return fixpoint of f.

    Uses laziness to avoid infinite recursion.
    The call fix(f) will result in f(Lazy(lambda: fix(f))).
    Thus, the argument of f is made lazy.  TODO: Why "make fix lazy"?  Terminology needs reconsideration.
    """
    f = func(f)
    # return f(Lazy(lambda: fix(f)))
    return f(lazy(fix)(f))


class Functor[X](Protocol):
    """Type class for functors.

    Deprecated (for use with (co)inductive types).
    """
    def __fmap__[Y](self, f: Callable[[X], Y]) -> "Functor[Y]":
        """Apply f inside self.
        """


@func
def fmap[F, X, Y](f: Callable[[X], Y]) -> "Callable[[F[X]], F[Y]]":
    """Apply functor F to function f in X -> Y to get function F[X] -> F[Y].

    Deprecated (for use with (co)inductive types).
    """
    @func(name=f"fmap({f})")
    def fmf(fx: "F[X]") -> "F[Y]":
        """Define F[f].
        """
        return fx.__fmap__(f)

    return fmf


# How a functor operates on functions
type Fmap[F, X, Y] = "Callable[ [Callable[[X], Y]], Callable[[F[X]], F[Y]] ]"


# type Iter[F: Functor, A] = A | F[Iter[F, A]]  # not accepted


@dataclass
class Fix[F: Functor]:
    """Fix[F] is fixpoint data type for functor F.

    For a lazy Fix, invoke as Fix(Lazy(lambda: v)).
    An alternative would be to define locally:

    .. code-block:: python

        class LazyFix[F, X](Fix[F[X]]):
            def unFix(self) -> F[Fix[F]]:
                return v

        ... LazyFix(None) ...

    """
    _unFix: "Final[F[Fix[F]] | Lazy[F[Fix[F]]]]"

    def unFix(self) -> "F[Fix[F]]":
        return evaluate(self._unFix)

    def __fmap__[G](self, f: Callable[[F], G]) -> "Callable[[Fix[F]], Fix[G]]":
        return Fix(f(self.unFix()))


@func(name="in")
def in_[F: Functor](a: "F[Fix[F]]") -> Fix[F]:
    """Constructor F-algebra of Fix[F].

    Deprecated type bound Functor.
    """
    return Fix(a)


@func
def out[F: Functor](a: Fix[F] | Lazy[Fix[F]]) -> "F[Fix[F]]":
    """Deconstructor F-coalgebra of Fix[F].

    Inverse of in.
    Deprecated type bound Functor.
    """
    return evaluate(a).unFix()  # TODO: why is evaluate needed?


type F_Algebra[F: Functor, X] = Callable[[F[X]], X]
# Deprecated type bound Functor.


@func(name="⦇ … ⦈_F")
def cata[F, X, Y](fmap: Fmap[F, X, Y]
                  ) -> "Callable[[F_Algebra[F, X]], Callable[[Fix[F]], X]]":
    """Construct catamorphism given functor and F-algebra (curried).
    """
    # below, lazy(rec) is needed in case the cata is applied to an infinite input
    # see prelude_test.py, test_cata_on_infinite_value()
    # N.B. This does mean that the algebra must call evaluate on the recursive input(s)
    # when this is not done automatically while processing infinite structures.
    return Func(lambda alg: (fix(lambda rec: alg @ fmap(lazy(rec)) @ out)
                             ).rename(f"⦇ {alg} ⦈_{fmap}"),
                name=f"⦇ … ⦈_{fmap}")


@func(name="⦇ … ⦈")
def cata_[F: Functor, X](alg: F_Algebra[F, X]) -> Callable[[Fix[F]], X]:
    """For F-algebra alg, construct catamorphism in Fix[F] -> X.

    Deprecated.
    """
    # return fix(lambda rec: alg @ fmap(rec) @ out)  # efficiency?
    @func(name=f"⦇ {alg} ⦈")
    def rec(a: Fix[F]) -> X:
        """rec = alg o F.rec o out"""
        return alg(fmap(rec)(out(a)))  # lazyf(rec) needed?
        # return (alg @ fmap(rec) @ out)(a)  # less efficient?

    return rec


type F_CoAlgebra[F: Functor, X] = Callable[[X], F[X]]
# Deprecated type bound Functor.


@func(name="〖 … 〗_F")
def ana[F, X, Y](fmap: Fmap[F, X, Y]
                 ) -> "Callable[[F_CoAlgebra[F, X]], Callable[[X], Fix[F]]]":
    """Construct anamorphism given functor and F-coalgebra (curried).
    """
    # below, lazy(rec) is needed in case the ana produces an (potentially) infinite result
    # N.B. lazy(fmap(rec)) is not good enough
    return Func(lambda coalg: (fix(lambda rec: in_ @ fmap(lazy(rec)) @ coalg)
                               ).rename(f"〖 {coalg} 〗_{fmap}"),
                name=f"〖 … 〗_{fmap}")


@func(name="〖 … 〗")
def ana_[F: Functor, X](coalg: F_CoAlgebra[F, X]) -> Callable[[X], Fix[F]]:
    """For F-coalgebra coalg, construct anamorphism in X -> Fix[F].

    Laziness needs to be built into the functor (?).
    Deprecated. ?
    """
    # return fix(lambda rec: in_ @ fmap(rec) @ coalg)  # not lazy; efficiency?
    @func(name=f"〖 {coalg} 〗")
    def rec(x: X) -> Fix[F]:
        """rec = in o F.rec o coalg"""
        return in_(fmap(rec)(coalg(x)))
        # return (in_ @ fmap(rec) @ coalg)(x)  # less efficient

    return rec


@func(name="⦇ … ⦈_F ⚬〖 … 〗_F")
def hylo[F, X, Y](fmap: Fmap[F, X, Y]
                ) -> "Callable[[F_Algebra[F, X], F_CoAlgebra[F, X]], Callable[[X], Y]]":
    """Construct hylomorphism given functor and F-(co)algebras (curried).
    """
    return lambda alg, coalg: Func(fix(lambda rec: alg @ fmap(lazy(rec)) @ coalg),
                                   name=f"⦇ {alg} ⦈ ⚬〖 {coalg} 〗")



"""
Definitions of basic functions for natural numbers and (potentially infinite) lists.

Copyright (c) 2024 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""
from typing import Any, Final, Callable
from FuPy.core import *

_ = x_  # even shorter name for the dummy argument


__all__ = [
    "not_", "and_", "or_",
    "swap", "coswap", "assoc_left", "assoc_right",
    "Maybe", "nothing", "just",
    "NatF", "natf", "Nat",
    "zero", "succ", "one",
    "NatF_Algebra", "NatF_CoAlgebra",
    "cataNat", "anaNat",
    "is_zero", "pred", "eqNat", "less",
    "intNat", "nat",
    "add", "mul", "sub", "div",
    "in_nat", "out_nat", "cata_nat", "ana_nat",
    "infinite",
    "ListF", "listf", "List_",
    "nil", "cons",
    "ListF_Alg", "ListF_CoAlg",
    "cataList", "anaList",
    "null", "head_tail", "head", "tail",
    "eqList",
    "strList", "listList", "list_",
    "length", "length_nat", "cat",
    "map_", "filter_",
    "concat",
    "take_drop", "take", "drop",
    "unzip", "zip", "zip_with",
    "from_",
    "null_", "cons_", "head_", "tail_", "cata_list", "ana_list",
]

# Functions on booleans

not_ = Func(lambda x: not x, name="¬")
and_ = Func(lambda x, y: x and y, name="∧")
or_ = Func(lambda x, y: x or y, name="∨")


# type: Func[Both[A, B], Both[B, A]]
swap = (second & first).rename("swap")

# type: Func[Either[A, B], Either[B, A]]
coswap = (right | left).rename("coswap")

# type: Func[Both[A, Both[B, C]], Both[Both[A, B], C]]
assoc_left = ((id_ * first) & second @ second).rename("assoc_left")  # (a, (b, c)) -> ((a, b), c)

# type: Func[Both[Both[A, B], C], Both[A, Both[B, C]]]
assoc_right = (first @ first & (second * id_)).rename("assoc_right")  # ((a, b), c) -> (a, (b, c))


type Maybe[X] = Either[Unit, X]  # 1 + X

nothing: Maybe[Any] = left(unit)  # inject left in Maybe[A]


@func
def just[A](a: A) -> Maybe[A]:
    """Inject right into Maybe[A]
    """
    return right(a)


@func
def maybef[X, Y](f: Callable[[X], Y]) -> Callable[[Maybe[X]], Maybe[Y]]:
    """Map f over a Maybe value.
    """
    return id_ + f


# (co)inductive type of natural numbers

type NatF[X] = Maybe[X]  # 1 + X


@func
def natf[X, Y](f: Callable[[X], Y]) -> Callable[[NatF[X]], NatF[Y]]:
    """Functor for type of natural numbers.
    """
    return id_ + f  # maybef(f)


# Alternative, defining NatF as instance of Functor
# @dataclass
# class NatF_[X](Functor[X]):
#     unNatF: Maybe[X]
#
#     @override
#     def __fmap__[Y](self, f: Callable[[X], Y]) -> "Functor[Y]":
#         return maybef(f)
#         return id_ + f  # natf(f), but the latter would add another indirection

# TODO: How to make a dataclass an instance of a protocol, afterwards?
# i.e., implement the required operations


# @func  # used in cataNat_
# def unNatF_[X](nfx: NatF_[X]) -> Maybe[X]:
#     return nfx.unNatF


type Nat = Fix[NatF]
# type Nat_ = Fix[NatF_]


zero: Final[Nat] = in_(left(unit))


@func(name="(+1)")
def succ(n: Nat) -> Nat:
    return in_(right(n))


one: Final[Nat] = succ(zero)


type NatF_Algebra[X] = F_Algebra[NatF, X]
type NatF_CoAlgebra[X] = F_CoAlgebra[NatF, X]


# TODO: Why not just define cataNat = cata(natf)?
@func(name=f"⦇ … ⦈_ℕ")
def cataNat[X](alg: NatF_Algebra[X]) -> Callable[[Nat], X]:
    """Catamorphism from NatF-algebra.
    """
    # return cata(natf)(alg)
    return cata(natf)(alg).rename(f"⦇ {alg} ⦈_ℕ")


# TODO: Why not just define cataNat_ = cata_ @ (_ @ unNatF_)?
# @func(name=f"⦇ … ⦈_ℕ_")
# def cataNat_[X](alg: F_Algebra[NatF_, X]) -> Callable[[Nat], X]:
#     """Catamorphism from NatF-algebra.
#     """
#     return cata_(alg @ unNatF_).rename(f"⦇ {alg} ⦈_ℕ_")


# TODO: Why not just define anaNat = ana(natf)?
@func(name=f"〖 … 〗_ℕ")
def anaNat[X](coalg: NatF_CoAlgebra[X]) -> Callable[[X], Nat]:
    """Anamorphism from NatF-coalgebra.
    """
    return ana(natf)(coalg).rename(f"〖 {coalg} 〗_ℕ")


# TODO: Why not just define anaNat_ = ana_ @ (NatF_ @ _)?
# @func(name=f"〖 … 〗_ℕ_")
# def anaNat_[X](coalg: F_CoAlgebra[X, NatF_]) -> Callable[[X], Nat]:
#     """Anamorphism from NatF-coalgebra.
#     """
#     return ana_(NatF_ @ coalg).rename(f"〖 {coalg} 〗_ℕ_")


# type: Callable[[Nat], bool]
# is_zero = (( const(True) | const(False) ) @ out).rename("is_zero")
is_zero = cataNat( const(True) | const(False) ).rename("is_zero")

# type: Callable[[Nat], Nat]  # partial function
pred = (( undefined | id_ ) @ out).rename("pred")
# tuple with identity
# pred = first @ cataNat( const(undefined, 0) | second & succ )

# type: Callable[[Nat], Callable[[Nat], bool]]
eqNat = cataNat( const(is_zero) | (lambda eq_n: (const(False) | eq_n @ pred) @ guard(is_zero)) ).rename("eqNat")
less = cataNat( const(not_ @ is_zero) | (lambda less_n: (const(False) | less_n @ pred) @ guard(is_zero)) ).rename("less")

# convert Nat to int

# type: Callable[[Nat], int]
intNat = cataNat( const(0) | (_ + 1) ).rename("intNat")

# convert int >= 0 to Nat
# type: Callable[[int], Nat]
nat = anaNat( (const(unit) + (_ - 1)) @ guard((_ == 0)) ).rename("nat")

#  various binary operators on Nat
# type: Callable[[Nat], Callable[[Nat], Nat]]
add = Func(lambda m: cataNat( const(m) | succ ), name="add")
mul = Func(lambda m: cataNat( const(zero) | add(m) ), name="mul")
sub = Func(lambda m: cataNat( const(m) | pred ), name="sub")
div = Func(lambda m: lambda n: anaNat( ( const(unit) + flip(sub)(n) ) @ guard(flip(less)(n)) )(m),
           name="div")

infinite: Final[Nat] = div(one)(zero)

# (co)inductive nat type based on native int

# TODO: use these in intNat and nat (?)
in_nat = (const(0) | (_ + 1)).rename("in_nat")
out_nat = ((const(unit) + (_ - 1)) @ guard((_ == 0))).rename("out_nat")


@func
def cata_nat[X](alg: Callable[[NatF[X]], X]) -> Func[int, X]:
    return fix(lambda rec: alg @ natf(rec) @ out_nat).rename(f"⦇ {alg} ⦈_nat")


@func
def ana_nat[X](coalg: Callable[[X], NatF[X]]) -> Func[X, int]:
    return fix(lambda rec: in_nat @ natf(rec) @ coalg).rename(f"〖 {coalg} 〗_nat")


# (co)inductive list type

type ListF[X] = Either[Unit, Both[Any, X]]  # 1 + A x X


@func
def listf[X, Y](f: Callable[[X], Y]) -> Callable[[ListF[X]], ListF[Y]]:
    """Functor for list type.
    """
    return id_ + id_ * f


type List_ = Fix[ListF]


nil: Final[List_] = in_(left(unit))


@func
def cons(x: Any, xs: List_) -> List_:
    """Uncurried operator to combine head and tail.
    """
    return in_(right((x, xs)))


type ListF_Alg[X] = F_Algebra[ListF, X]
type ListF_CoAlg[X] = F_CoAlgebra[ListF, X]


@func
def cataList[X](alg: ListF_Alg[X]) -> Callable[[List_], X]:
    """Catamorphism from ListF-algebra.
    """
    return cata(listf)(alg).rename(f"⦇ {alg} ⦈_𝕃")


@func
def anaList[X](coalg: ListF_CoAlg[X]) -> Callable[[X], List_]:
    """Anamorphism from ListF-algebra.
    """
    return ana(listf)(coalg).rename(f"〖 {coalg} 〗_𝕃")


# type: Callable[[List_], bool]
null = cataList( const(True) | const(False) ).rename("null")
# null = ( const(True) | const(False) ) @ out

# type: Callable[[List_], Any]
head_tail = (( undefined | id_ ) @ out).rename("head_tail")

# type: Callable[[List_], Any]
head = (first @ head_tail).rename("head")

# type: Callable[[List_], List_]
tail = (second @ head_tail).rename("tail")

# check whether two List_'s are equal (curried)
# type: Callable[[Nat], Callable[[Nat], bool]]
eqList = cataList( const(null)
                 | Func(lambda x, r:  # r = eqList(xs)
                              ( const(False)
                              | and_ @ (Func(lambda y: x == y) * r) @ head_tail
                              ) @ guard(null))
                 )

# represent a List_ as a string
# type: Callable[[List_], str]
strList = cataList(const("[]") | Func(lambda x, xs: f"{x} ⊢ {evaluate(xs)}"))

# convert a List_ to a regular Python list
# type: Callable[[List_], list]
listList = cataList(const([]) | Func(lambda x, xs: [x] + xs))

# convert regular Python list to a List_
# type: Callable[[list], List_]
list_ = anaList( (const(unit) + (Func(lambda xs: xs[0]) & (lambda xs: xs[1:]))) @ guard(not_) )

length = cataList(const(zero) | succ @ second).rename("length")
length_nat = cataList(const(0) | (1 + _) @ second).rename("length_nat")


@func
def cat(xs: List_, ys: List_) -> List_:
    """Uncurried catenation on List_.
    """
    return cataList( const(ys) | cons )(xs)


# map function over list
# type: Callable[[Callable[[X], Y]], Callable[[List_], List_]]
map_ = Func(lambda f: cataList( const(nil) | cons @ (f * id_) ).rename(f"map({f}"),
            name="map")

# filter list by predicate
# type: Callable[[Callable[[X], bool]], Callable[[List_], List_]]
filter_ = Func(lambda p: cataList(const(nil) | (cons | second) @ guard(p @ first)).rename(f"filter({p})"),
               name="filter")

concat = cataList( const(nil) | cat ).rename("concat")

take_drop = cata_nat( const(const(nil) & id_) |
                      Func(lambda tdn: (const(nil, nil) | (cons * id_) @ assoc_left @ (id_ * tdn) @ head_tail) @ guard(null))
                      ).rename("take_drop")
take = Func(lambda n: first @ take_drop(n)).rename("take")
drop = Func(lambda n: second @ take_drop(n)).rename("drop")
# drop = cata_nat( const(id_) | (lambda r: (const(nil) | r @ tail) @ guard(null))
#                  ).rename("drop")
# take = cata_nat( const(const(nil)) | (lambda r: (const(nil) | cons @ (id_ * r) @ head_tail) @ guard(null))
#                  ).rename(f"take")

unzip = cataList(const(nil, nil) | Func(lambda xy, ts: (cons(first(xy), first(ts)), cons(second(xy), second(ts))))).rename("unzip")
zip = anaList( (const(unit) + ((head * head) & (tail * tail))) @ guard(Func(lambda xs, ys: null(xs) or null(ys))) ).rename("zip")
zip_with = Func(lambda op: anaList( (const(unit) + (op @ (head * head) & (tail * tail))) @ guard(Func(lambda xs, ys: null(xs) or null(ys))) )
                ).rename("unzip_with")

from_ = anaList( right @ (id_ & (_ + 1)) ).rename("from")

# (co)inductive list_ type based on native list (as snoc lists)

null_ = la('xs: not xs').rename("null_")
cons_ = la('(x, xs): [x] + xs').rename("cons_")
head_ = la('xs: xs[0]').rename("head_")
tail_ = la('xs: xs[1:]').rename("tail_")
in_list = (const([]) | cons_).rename("in_list")
out_list = ((const(unit) + (head_ & tail_)) @ guard(null_)).rename("out_list")


@func
def cata_list[X](alg: Callable[[ListF[X]], X]) -> Func[list, X]:
    # return fix(lambda rec: alg @ listf(rec) @ out_list).rename(f"⦇ {alg} ⦈_list")
    return fix(la('rec: alg @ listf(rec) @ out_list')).rename(f"⦇ {alg} ⦈_list")
    # for infinite arguments, use lazy(rec)


@func
def ana_list[X](coalg: Callable[[X], ListF[X]]) -> Func[X, list]:
    return fix(lambda rec: in_list @ listf(rec) @ coalg).rename(f"〖 {coalg} 〗_list")
    # for infinite results, use lazy(rec)



"""
Definitions to support functional programming in Python.
Especially aimed at educational use and not at industrial application.

Copyright (c) 2024 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""
import collections.abc as abc
import inspect
import FuPy.tracing as tr
from dataclasses import dataclass
from functools import update_wrapper, wraps
from typing import Any, Final, Optional, Callable, Protocol, override

from FuPy import utils
from FuPy.tracing_basics import *

__all__ = [
    "Empty", "Unit", "unit",
    "Either", "Both",
    "Func", "func", "la",
    "undefined", "eq",
    "id_", "const", "left", "right", "guard", "first", "second",
    "curry", "uncurry", "flip", "ap",
    "compose", "split", "case_", "ftimes", "fplus", "fpower",
    "fpower_left",
    "x_",
]


class Empty:
    def __new__(cls, *args, **kwargs):
        raise TypeError("Empty cannot be instantiated")


class Unit:
    """The Unit type 𝟙 with only the unit value.

    A singleton class.
    """
    _has_instance = False

    def __init__(self) -> None:
        """Enforce singleton.
        """
        if Unit._has_instance:
            raise RuntimeError("Unit is a singleton class; use `unit` instead")
        Unit._has_instance = True

    @override
    def __repr__(self) -> str:
        return "unit"

    @override
    def __str__(self) -> str:
        return "_"  # 2IPH0 math notation

    def __len__(self) -> int:
        return 0

    def __iter__(self):
        """Makes unit unpackable.
        Also see __rmatmul__ (@) in Func.
        """
        return iter(())


unit: Final[Unit] = Unit()  # the unit value
# We could also have used () as unit value,
# but then we could not type hint it with tuple[]
# Also, then we could not give it a custom str/repr.


@dataclass
class Left[A]:
    """The left option of the sum type Either.
    """
    value: Final[A]


@dataclass
class Right[B]:
    """The right option of the sum type Either.
    """
    value: Final[B]


# Sum type A + B is Either[A, B]
type Either[A, B] = Left[A] | Right[B]


# Product type A x B is Both[A, B]
type Both[A, B] = tuple[A, B]


class Func[A, B]:
    """Wrapper for functions A -> B.

    Supports various operations on functions:

    - application (:code:`f(arg)`), composition (:code:`f @ g`)
    - split (:code:`f & g`), case (:code:`f | g`)
    - product (:code:`f * g`), sum (:code:`f + g`)
    - iteration (:code:`f ** n`)

    These operators can be sectioned.

    It would be nice if this could be incorporated into types.FunctionType
    (which is immutable; so, these cannot be added dynamically.)

    Note that in an FP language like Haskell, functions take exactly one argument
    (and return one result).
    That is why we give a Func the unit value as default argument.
    There is some tension between this view and the broader view of Python on functions
    (the latter can have no arguments or multiple arguments).
    The Func wrapper handles this by automatically (un)packing of arguments.
    Consider the call Func(f)(*args) where args is a tuple and
    f is defined as `def f(*fargs): ...` where fargs is also a tuple.
    Let R be the number of _required_ non-keyword arguments of f
    and :code:`A =  len(args)` the number of arguments of f.
    The call :code:`Func(f)(*args)` is handled as follows:

    +--------+-----------+-------------------+--------------+
    |        | A == 0    | A == 1            | A >= 2       |
    +========+===========+===================+==============+
    | R == 0 | f(*args)  | f(*(args[0])) (a) | TypeError    |
    +--------+-----------+-------------------+--------------+
    | R == 1 | f(unit)   | f(args[0])        | f(args) (b)  |
    +--------+-----------+-------------------+--------------+
    | R >= 2 | TypeError | f(*(args[0])) (c) | f(*args) (d) |
    +--------+-----------+-------------------+--------------+

    Notes:

    * (a) args[0] must behave like an empty tuple, e.g. unit.
    * (b) pack the arguments and pas them as a single tuple.
    * (c) unpack the argument and pass the unpacked values separately;
          N.B. the single argument must be an iterable of appropriate length
    * (d) the number of arguments provided and expected must match.
    * Keyword parameters are passed on.
    """
    COMBINATORS = {'@': '⚬', '&': '△', '|': '▽', '*': '⨯', '+': '+'}
    PRIORITIES = {'': 9, '.': 8, '^': 7, '@': 5, '&': 3, '*': 3, '|': 1, '+': 1}
    fully_parenthesize = False  # affects self.str()

    def __init__(self, f: Callable[..., B], name: Optional[str] = None,
                 r: Optional[int] = None, top: str = '') -> None:
        """Wrap f in Func.
        * self.name is how it prints in math notation.
        * self.r is minimum number of required arguments
          (None means: attempt to determine r from f, defaulting to 1)
        * self.top encodes top-level operator in f's construction
        """
        if r is None:
            try:
                r = utils.count_required_args(f)
            except ValueError as e:
                r = 1  # default for functions that cannot be inspected
        self.required_args_count = r
        self.func = f
        self.top = top  # top-level operator, used to suppress unneeded parentheses
        update_wrapper(self, f, updated={})
        self.name = name or (f.__name__ if hasattr(f, '__name__') else str(f))

    @override
    def __repr__(self) -> str:
        return f"Func({self}, r={self.required_args_count}, top={self.top!r})"

    @override
    def __str__(self) -> str:
        return self.name

    def str(self, op: str) -> str:
        """Return name in the context of operator op, adding parentheses if needed.
        """
        if Func.fully_parenthesize:
            if self.top not in ('', '^') or op == '':
                return f"({self})"
        if Func.PRIORITIES[self.top] > Func.PRIORITIES[op]:
            return str(self)
        elif self.top == op and op in ('^', '@'):
            return str(self)
        else:
            return f"({self})"

    def show_top(self) -> str:
        """Show top-level function for motivation of reduction.
        """
        if self.top == '':
            if self.name.startswith('(λ'):
                return self.name[1]
            return self.name
        elif self.top == '^':
            if self.name.endswith(('°', '?')):
                return self.name[-1]
            return '^'
        else:
            return Func.COMBINATORS[self.top]

    def rename(self, name: str, top: Optional[str] = None) -> "Func[A, B]":
        """Rename f.
        """
        self.name = name
        if top is not None:
            self.top = top
        return self

    def define_as(self, name: str) -> "Func[A, B]":
        """Define name as self.
        """
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(self, name=name, top='')

    def doc(self, docstring: str) -> "Func[A, B]":
        """Add docstring to self.
        """
        self.__doc__ = docstring
        return self

    def __call__(self, *args: A, **kwargs) -> B:
        """Apply self to args.
        """
        # TODO: Should keyword arguments be passed as well?
        actual_args_count = len(args)
        # print(f"{self} requires {self.required_args_count} args; provided args: {args}")
        # print(f"call entry: {self.str('^')}{args!r}")
        # Cases to distinguish
        # R = self.required_args_count
        # R == 0 and args == (): treat as unit; new_args = args; show as (unit)
        #        and args == (x,): check that x == unit or x == (); new_args = (); show as args
        #        and args == (x, y, ...): TypeError
        # R == 1 and args == (): new_args = (unit,); show as new_args
        #        and args == (x,): new_args = args; show as either, but when isinstance(x, tuple), show as x
        #        and args == (x, y, ...): new_args = (args,); auto pack; show as args
        # R >= 2 and args == (): TypeError
        #        and args == (x,): new_args = x; auto unpack; show as x
        #        and args == (x, y, ...): new_args = args; show as either
        comment = ""
        if self.required_args_count == 1:  # pass required one argument
            if actual_args_count == 1:  # "normal" 1-from-1: pass on as is
                new_args = args
                if isinstance(args[0], tuple):
                    show_args_as = args[0]  # avoid double parentheses around arguments
                else:
                    show_args_as = args
            elif not actual_args_count:  # the 1-from-0 case
                new_args = (unit,)
                show_args_as = new_args
                comment = "; auto-provide unit argument"
            else:  # 1-from-2+: pass as tuple (auto-pack)
                new_args = (args,)
                show_args_as = args
                comment = "; auto-pack arguments in tuple"
        else:  # f requires zero or more than one argument
            if actual_args_count == 1:  # x-from 1 where x != 1: auto-unpack the 1 argument
                new_args = args[0]  # it must be iterable and of matching length (checked below)
                if not self.required_args_count:  # 0-from-1: auto convert unit
                    if new_args is not unit and new_args != ():
                        raise TypeError(f"{self.func!r} requires no argument, but actual argument {new_args!r} is not unit")
                    show_args_as = (unit,)
                    comment = "; auto-provide unit argument"
                else:  # self.required_args_count >= 2
                    if not isinstance(new_args, tuple):
                        raise TypeError(f"{self.func!r} requires {self.required_args_count} arguments, "
                                        f"but actual argument {new_args!r} is not a tuple")
                    show_args_as = new_args
                    comment = "; auto-unpack arguments from tuple"
            else:  # x-from-y where y != 1: pass on as is (numbers must match; checked below)
                new_args = args
                show_args_as = args
        # print(f"  passed args: {new_args}")

        # TODO: in trace step, mention whether evaluation is lazy
        tr.trace_step(lambda: ApplicationStep(f"{self.str('.')}.{utils.show_args(show_args_as)}"))
        tr.trace_step(lambda: MotivationStep(f"def. of {self.show_top()}{comment}"))

        actual_args_count = len(tuple(new_args))
        if self.required_args_count != actual_args_count:
            raise TypeError(f"{self.func!r} argument mismatch: {self.required_args_count} required, "
                            f"but {actual_args_count} given")
        tr.inc_depth()
        # TODO: full laziness; but this approach does not work.
        # from FuPy.laziness import Lazy, evaluate
        # if (Lazy.full_regime and
        #         (isinstance(arg, Lazy) for arg in new_args) and
        #         self.show_top() != 'λ' and
        #         self is not evaluate
        #     ):
        #     print("PROPAGATING LAZINESS")
        #     result = Lazy(la(': self.func(*(evaluate(arg) for arg in args), **kwargs)'),
        #                   name=f"(λ: {self.name}{utils.show_args(new_args)})")
        # else:
        result = self.func(*new_args, **kwargs)
        # print(f"  result: {result}")
        # print(f"call exit: {self.str('^')}{new_args!r} ~> {result!r}")
        tr.dec_depth()
        tr.trace_step(lambda: ResultStep(f"{utils.show_value(result)}"))
        return result

    def __matmul__[Z](self, g: Callable[[Z], A]) -> "Func[Z, B]":
        """Return self o g.

        If g is not an instance of Func, then it is wrapped as Func.
        The typing is difficulty to express:
        * if g expects no or more than one argument,
          then the composition has a single tuple argument.
        """
        if not isinstance(g, abc.Callable):
            return NotImplemented
        g = func(g)
        op = " ⚬ " if self.required_args_count == 1 else "* ⚬ "
        name = f"{self.str('@')}{op}{g.str('@')}"
        # return Func(lambda z: self(g(z)),
        #             name=f"{self.str('@')}{op}{g.str('@')}", top='@')
        # print(f"constructing {name}")
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda z: self(g(z)), name, top='@')
        # return la('z: self(g(z))').rename(name, top='@')

    def __rmatmul__[C](self, g: Callable[..., C]) -> "Func[A, C]":
        """Return g o self.
        """
        if not isinstance(g, abc.Callable):
            return NotImplemented
        return func(g) @ self

    def __and__[C](self, g: Callable[[A], C]) -> "Func[A, Both[B, C]]":
        """Return self split g.
        """
        if not isinstance(g, abc.Callable):
            return NotImplemented
        g = func(g)
        name = f"{self.str('&')} △ {g.str('&')}"
        # print(f"constructing {name}")
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda a: (self(a), g(a)), name=name, top='&')

    def __rand__[C](self, g: Callable[[A], C]) -> "Func[A, Both[C, B]]":
        """Return g split self.
        """
        if not isinstance(g, abc.Callable):
            return NotImplemented
        return func(g) & self

    def __or__[Z](self, g: Callable[[Z], B]) -> "Func[Either[A, Z], B]":
        """Return self case g.
        """
        if not isinstance(g, abc.Callable):
            return NotImplemented
        g = func(g)
        name = f"{self.str('|')} ▽ {g.str('|')}"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))

        @func(name=name, top="|")
        def case_(x: Either[A, Z]) -> B:
            match x:
                case Left(value=a):
                    return self(a)
                case Right(value=z):
                    return g(z)
                case _:
                    raise TypeError(f"type of argument {x!r} of {name} is neither Left nor Right")

        return case_

    def __ror__[Z](self, g: Callable[[Z], B]) -> "Func[Either[Z, A], B]":
        """Return g case self.
        """
        if not isinstance(g, abc.Callable):
            return NotImplemented
        return func(g) | self

    def __mul__[C, D](self, g: Callable[[C], D]) -> "Func[Both[A, C], Both[B, D]]":
        """Return self x g.
        """
        if not isinstance(g, abc.Callable):
            return NotImplemented
        g = func(g)
        name = f"{self.str('*')} ⨯ {g.str('*')}"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda a, c: (self(a), g(c)),  # alternative: self @ first & g @ second
                    name=name, top='*')

    def __rmul__[C, D](self, g: Callable[[C], D]) -> "Func[Both[C, A], Both[D, B]]":
        """Return g x self.
        """
        if not isinstance(g, abc.Callable):
            return NotImplemented
        return func(g) * self

    def __add__[C, D](self, g: Callable[[C], D]) -> "Func[Either[A, C], Either[B, D]]":
        """Return self + g.
        """
        if not isinstance(g, abc.Callable):
            return NotImplemented
        g = func(g)
        name = f"{self.str('+')} + {g.str('+')}"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return (left @ self | right @ g).rename(name, top='+')

    def __radd__[C, D](self, g: Callable[[C], D]) -> "Func[Either[C, A], Either[D, B]]":
        """Return g + self.
        """
        if not isinstance(g, abc.Callable):
            return NotImplemented
        return func(g) + self

    def __pow__(self: "Func[A, A]", n: Any) -> "Func[A, A]":
        """Repeated composition.
        A.k.a. as function exponentiation.
        This is fold-right over nat.

        Assumptions: A == B and n >= 0.
        """
        if not isinstance(n, int):
            return NotImplemented
        if n == 0:
            return id_
        elif n == 1:  # optimization
            return self
        else:
            return (self @ self ** (n - 1)).rename(
                f"{self.str('^')}^{n}", top='^')
            # N.B. f ** -1 does not terminate
            # We can make f ** -1 = fix(f)
            # return (lazy(self ** _)(n - 1) @ self).rename(f"{self}^{n}")  # fold-left over nat
            # TODO: which is better:  self @ self ** (n - 1)  or  self ** (n - 1) @ self


def func(obj: Optional[Callable] = None, **kwargs) -> Callable:
    """Decorate obj by wrapping it in Func.

    When obj is None, it returns a wrapping function.
    Avoids double wrapping.
    When f is a class, it wraps obj.__init__ (though such functions cannot be used with `|`).
    """
    @wraps(func)
    def func_decorator(obj: Callable) -> Callable:
        if inspect.isclass(obj):
            if isinstance(obj.__init__, Func):  # don't decorate again
                return obj
            obj.__init__ = Func(obj.__init__, **kwargs)
            return obj
        else:  # a regular function
            if isinstance(obj, Func):  # don't decorate again
                return obj
            return Func(obj, **kwargs)

    if obj is None:
        return func_decorator
    else:
        return func_decorator(obj)


def la(la_expr: str) -> Func[Any, Any]:
    """Define a lambda abstraction with zero or more variables.
    Returns a curried or uncurried function.

    Argument :code:`la_expr` must have the shape:
        :code:`variables: expr`
    where variables are comma separated: variable, ..., variable,
    where each variable can be
    * empty (no name/tuple)
    * a single name
    * a tuple of zero or more names

    Empty or an empty tuple results in a constant function with unit as argument.
    A tuple of 2 or more names results in an uncurried function.
    Multiple variables result in a curried function.

    :code:`la('a, (b, c), , d: (a, b, c, d, e, f)')` where :code:`e == 42` locally and :code:`f` is global,
    results (roughly) in:
    .. code-block:: python

        Func(lambda a:
             Func(lambda b, c:  # this is an uncurried function
                  Func(lambda:  # this is a constant function
                       Func(lambda d: (a, b, c, d, e),
                            name=...),
                       name=...),
                  name=...),
             name=...)

    Recursive approach, with auto capturing of locals.
    """
    # TODO: investigate performance loss, due to parsing and eval
    # from pprint import pprint
    # pprint(f"la_expr: {la_expr}")
    try:
        parameters, body = la_expr.split(':', 1)
    except ValueError:
        raise SyntaxError(f"la({la_expr!r}) has no `:`")
    parameters, body = parameters.strip(), body.strip()
    # find first parameter (could be more robust?)
    if parameters.startswith('('):  # tuple of parameters
        try:
            parameter, rest = parameters.lstrip('(').split(')', 1)  # rest is empty if ')' not present
        except ValueError:
            raise SyntaxError(f"la({la_expr!r}) has `(` without `)`")
        rest = rest.split(',', 1)[1:]
    else:  # single parameter
        parameter, *rest = parameters.split(',', 1)
    # N.B. parameter can be empty
    # len(rest) is 1 or 0, depending on whether there was a comma or not

    parent_frame = inspect.currentframe().f_back
    parent_globals = parent_frame.f_globals
    parent_locals = parent_frame.f_locals
    # pprint(f"parent_locals: {parent_locals}")
    try:
        lambda_body = eval(f"lambda: {body}")
    except SyntaxError as excinfo:
        raise SyntaxError(f"la({la_expr!r}) body `{body}`: {excinfo}")
    # pprint(f"names: {lambda_body.__code__.co_names}")
    if parent_locals is parent_globals:
        custom_locals = {}
    else:
        custom_locals = {key: parent_locals[key]
                         for key in lambda_body.__code__.co_names
                         if key in parent_locals}
    # pprint(f"custom_locals: {custom_locals}")
    bindings = '\n'.join(utils.indent_lines(f"{var} = {utils.show_value(value)}", '  ')
                         for var, value in custom_locals.items())
    # pprint(bindings)
    where_clause = f" where {{\n{bindings}\n}}" if bindings else ''
    closure = ','.join(f" {var}={var}"
                       for var in custom_locals)
    # pprint(closure)
    new_body = f"la('{rest[-1].strip()}: {body}')" if rest else body  # TODO: what if body contains quotes?
    la_expr_new = f"lambda{' ' if parameter else ''}{parameter}{',' if closure and parameter else ''}{closure}: {new_body}"
    # pprint(f"la_expr_new: {la_expr_new}")
    name=f'''(λ {parameters}: `{body}`){where_clause}'''
    tr.trace_step(lambda: DefinitionStep(f"{name}"))
    try:
        new_func = eval(la_expr_new, parent_globals, custom_locals)
    except SyntaxError as excinfo:
        raise SyntaxError(f"la({la_expr!r}) has bad parameter `{parameter}`: {excinfo}")
    return Func(new_func, name=name)


@func(name="⊥", r=1)
def undefined[A](a: A = unit) -> Empty:
    """Abort evaluation.
    """
    raise ValueError(f"undefined({'' if a is unit else repr(a)})")


@func
def eq[A](a: A, b: A) -> bool:
    """Uncurried equality test.
    """
    return a == b


@func(name="id")
def id_[A](a: A) -> A:
    """Polymorphic identity function.
    """
    return a


@func(name="(°)")
def const[A, B](b: B) -> Func[A, B]:
    """Construct the constant function always returning b.
    """
    return Func(lambda a: b, name=f"{b.str('^') if isinstance(b, Func) else b}°", top='^')


@func(name="↪︎")
def left[A, B](a: A) -> Either[A, B]:
    """Inject left into sum type.
    """
    return Left(a)


@func(name="↩︎")
def right[A, B](b: B) -> Either[A, B]:
    """Inject right into sum type.
    """
    return Right(b)


@func(name="(?)")
def guard[A](p: Callable[[A], bool]) -> Callable[[A], Either[A, A]]:
    """For predicate p, construct function p? in (A -> bool) -> A -> Either[A, A].
    """
    @func(name=f"{p.str('^')}?", top='^')
    def aux(a: A) -> Either[A, A]:
        """p?(a)"""
        if p(a):
            return Left(a)
        else:
            return Right(a)

    return aux


@func(name="≪")
def first[A, B](t: Both[A, B]) -> A:
    """Project left out of product type.
    """
    return t[0]


@func(name="≫")
def second[A, B](t: Both[A, B]) -> B:
    """Project right out of product type.
    """
    return t[1]


def _parenth(a: Any) -> str:
    """Parenthesize a if needed.
    """
    if isinstance(a, Func):
        return a.str('^')
    else:
        return a


@dataclass
class OperatorSection:
    """Support for operator sectioning.
    OperatorSection() is used as dummy argument to implicitly create lambda expressions.
    It must not implement __call__, because otherwise an OperatorSection object will not work
    with Func function combinators as expected.
    """
    name: Final[str] = ''

    @override
    def __repr__(self) -> str:
        """Prints nothing, which is needed when both operands are omitted.
        """
        return f"OperatorSection(name={self.name!r})"

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other):
        name = f"(={other})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: x == other, name=name)

    def __ne__(self, other):
        name = f"(≠{other})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: x != other, name=name)

    def __ge__(self, other):
        name = f"(≥{other})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: x >= other, name=name)

    def __gt__(self, other):
        name = f"(>{other})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: x > other, name=name)

    def __le__(self, other):
        name = f"(≤{other})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: x <= other, name=name)

    def __lt__(self, other):
        name = f"(<{other})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: x < other, name=name)

    def __add__(self, other):
        name = f"(+{' ' if callable(other) else ''}{_parenth(other)})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: x + other, name=name)
        # return la('x: x + other').rename(name)  # TODO: consider

    def __radd__(self, other):
        name = f"({_parenth(other)}{' ' if callable(other) else ''}+)"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: other + x, name=name)

    def __sub__(self, other):
        name = f"(-{other})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: x - other, name=name)

    def __rsub__(self, other):
        name = f"({other}-)"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: other - x, name=name)

    def __mul__(self, other):
        name = f"(*{' ' if callable(other) else ''}{_parenth(other)})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: x * other, name=name)

    def __rmul__(self, other):
        name = f"({_parenth(other)}{' ' if callable(other) else ''}*)"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: other * x, name=name)

    def __matmul__(self, other):
        name = f"(⚬{' ' if callable(other) else ''}{_parenth(other)})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: x @ other, name=name)

    def __rmatmul__(self, other):
        name = f"({_parenth(other)}{' ' if callable(other) else ''}⚬)"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: other @ x, name=name)

    def __truediv__(self, other):
        name = f"(/{other})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: x / other, name=name)

    def __rtruediv__(self, other):
        name = f"({other}/)"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: other / x, name=name)

    def __floordiv__(self, other):
        name = f"(div {other})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: x // other, name=name)

    def __rfloordiv__(self, other):
        name = f"({other} div)"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: other // x, name=name)

    def __mod__(self, other):
        name = f"(mod {other})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: x % other, name=name)

    def __rmod__(self, other):
        name = f"({other} mod)"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: other % x, name=name)

    def __pow__(self, other):
        name = f"(**{other})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: x ** other, name=name)

    def __rpow__(self, other):
        name = f"({other}**)"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: other ** x, name=name)

    def __and__(self, other):
        name = f"({'△ ' if callable(other) else '&'}{_parenth(other)}"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: x & other, name=name)

    def __rand__(self, other):
        name = f"({_parenth(other)}{' △' if callable(other) else '&'})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: other & x, name=name)

    def __or__(self, other):
        name = f"({'▽ ' if callable(other) else '&'}{_parenth(other)})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: x | other, name=name)

    def __ror__(self, other):
        name = f"({_parenth(other)}{' ▽' if callable(other) else '&'})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: other | x, name=name)

    def __divmod__(self, other):
        name = f"divmod( , {other})"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: divmod(x, other), name=name)

    def __rdivmod__(self, other):
        name = f"divmod({other}, )"
        tr.trace_step(lambda: DefinitionStep(f"{name}"))
        return Func(lambda x: divmod(other, x), name=name)


# object.__divmod__(self, other)
# object.__lshift__(self, other)
# object.__rshift__(self, other)
# object.__xor__(self, other)
#
# object.__rdivmod__(self, other)
# object.__rlshift__(self, other)
# object.__rrshift__(self, other)
# object.__rxor__(self, other)


# Client code may want to define `_ = x_` to simplify operator section notation.
# This is not done here, because `_` has a special meaning in some REPLs.
x_ = OperatorSection()


@func
def curry[A, B, C](f: "Callable[[Both[A, B]], C]") -> "Func[A, Func[B, C]]":
    """Curry f.
    """
    return Func(lambda a: Func(lambda b: f((a, b)),
                               name=f"curry({f})({a})"),
                name=f"curry({f})")


@func
def uncurry[A, B, C](f: Callable[[A], Callable[[B], C]]) -> Func[Both[A, B], C]:
    """Uncurry f.
    """
    return Func(lambda t: f(first(t))(second(t)),
                name=f"uncurry({f})")


@func
def flip[A, B, C](f: Callable[[A], Callable[[B], C]]) -> Func[B, Func[A, C]]:
    """Flip f.  Swaps arguments on curried functions.
    """
    return Func(lambda b: Func(lambda a: f(a)(b), name=f"flip({f})({b})"), name=f"flip({f})")


# function operators as uncurried functions
ap = Func(lambda f, x: f(x), name="ap")

compose = Func(lambda f, g: f @ g, name="compose")  # composition
split = Func(lambda f, g: f & g, name="split")
# split = la('(f, g): f & g').rename("split")
case_ = Func(lambda f, g: f | g, name="case")
ftimes = Func(lambda f, g: f * g, name="ftimes")
fplus = Func(lambda f, g: f + g, name="fplus")
fpower = Func(lambda f, n: f ** n, name="fpower")


@func
def fpower_left[A](f: Func[A, A], n: int) -> Func[int, A]:
    """Repeated composition.
    This is fold-left over nat.
    It is tail recursive.
    See FuPy.laziness.fpower_lazy for a version that does not use the stack.

    Assumption: n >= 0
    """
    return id_ if n == 0 else fpower_left(f, n - 1) @ f



"""
Package FuPy - Functional Programming in Python

* FuPy.core loads FuPy.basics, FuPy.tracing, FuPy.laziness, FuPy.fixpoints
* FuPy.all loads FuPy.core and FuPy.prelude

Copyright (c) 2024 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""



""" tracing_laziness.py

Definitions to support tracing of laziness functionality.

Copyright (c) 2024 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""
from typing import override

import FuPy.tracing as tr

__all__ = [
    "SuspensionStep", "EvaluationStep", "GettingStep",
]


class SuspensionStep(tr.BaseStep):
    """Class for `Lazy` creation steps.
    """
    def __init__(self, note: str) -> None:
        super().__init__(f"{note}")

    @override
    def __str__(self) -> str:
        return f"Suspend:   {super().__str__()}"


class EvaluationStep(tr.BaseStep):
    """Class for evaluation steps.
    """
    def __init__(self, note: str) -> None:
        super().__init__(f"{note}")

    @override
    def __str__(self) -> str:
        return f"Evaluate:  {super().__str__()}"


class GettingStep(tr.BaseStep):
    """Class for getting steps (from cache).
    """
    def __init__(self, note: str) -> None:
        super().__init__(f"{note}")

    @override
    def __str__(self) -> str:
        return f"Get:       {super().__str__()}"



"""
Load all FuPy functionality.

Copyright (c) 2024 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""
from .core import *
from .prelude import *
#
# Sphinx needs __all__ for its automodule functionality
from .core import __all__ as core_all
from .prelude import __all__ as prelude_all

__all__ = core_all + prelude_all



"""
Definitions to support tracing of functional programs in Python.

Copyright (c) 2024 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""
from dataclasses import dataclass, field
from typing import Any, Callable, override, Optional
from abc import ABC, abstractmethod
from FuPy import utils

__all__ = [
    "Trace",
    "BaseStep",
    "trace_step", "inc_depth", "dec_depth", "trace",
]


@dataclass
class Trace:
    """A Trace object contains a sequence of evaluation (rewrite) steps.
    """
    trace: list["BaseStep"] = field(default_factory=list)
    depth: int = 0
    skip: set[type["BaseStep"]] = field(default_factory=set)  # suppressed steps in __str__ and live tracing

    # @override
    # def __repr__(self) -> str:
    #     return repr(f"trace={self.trace}")

    def __str__(self) -> str:
        return '\n'.join(f"{index:4} {step}" for index, step in enumerate(self.trace) if type(step) not in self.skip)

    def log(self, step: "BaseStep") -> None:
        """Append a step to the trace.
        """
        self.trace.append(step.set_depth(self.depth))

    def update_depth(self, delta: int) -> None:
        """Update the depth for the next step.
        """
        self.depth += delta


@dataclass
class BaseStep(ABC):
    """Abstract Base Class for all Trace steps.
    """
    note: str
    depth: int = 0

    @override
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(note={self.note!r}, depth={self.depth})"

    @override
    def __str__(self) -> str:
        indentation = self.depth * '│     '
        return f"{indentation}{utils.indent_lines(self.note, f"{16 * ' '}{indentation}", k=1)}"

    def set_depth(self, depth: int) -> "BaseStep":
        """Set the depth of the step.
        """
        self.depth = depth
        return self


# the global trace; assign Trace() to clear it and enable tracing
the_trace: Optional[Trace] = None
live_tracing: bool = False


def trace[A](expr: Callable[[], A],
             live: Optional[bool] = None,
             skip: Optional[set[type[BaseStep]]] = None) -> tuple[A, Trace]:
    """Trace evaluation of expr.
    The expression must be provided as a constant function,
    that is, in the form `lambda: expr`, or as `Lazy(lambda: expr)`.
    Later, possibly also as string.
    """
    global the_trace, live_tracing
    live_tracing_old = live_tracing
    the_trace = Trace()
    if live is not None:
        live_tracing = live
    if skip is not None:
        the_trace.skip = skip

    result = expr(), the_trace

    live_tracing = live_tracing_old
    return result


def trace_step(step_: Callable[[], BaseStep]) -> None:
    """Add step_() to the trace if tracing is enabled.
    For internal use only.
    """
    global the_trace, live_tracing
    if the_trace is None:
        return
    index = len(the_trace.trace)
    step = step_()
    the_trace.log(step)
    if live_tracing and type(step) not in the_trace.skip:
        print(f"{index:4} {step}")


def inc_depth() -> None:
    """Increment the tracing depth by one.
    """
    global the_trace
    if the_trace is not None:
        the_trace.update_depth(1)


def dec_depth() -> None:
    """Decrement the tracing depth by one.
    """
    global the_trace
    if the_trace is not None:
        the_trace.update_depth(-1)



"""
Core definitions to support functional programming in Python.
Especially aimed at educational and not at industrial application.

Copyright (c) 2024 - Eindhoven University of Technology, The Netherlands

This software is made available under the terms of the MIT License.
"""
from .basics import *
from .tracing import trace
from .laziness import *
from .fixpoints import *

# Sphinx needs __all__ for its automodule functionality
from .basics import __all__ as basics_all
from .laziness import __all__ as laziness_all
from .fixpoints import __all__ as fixpoints_all

__all__ = basics_all + ["trace"] + laziness_all + fixpoints_all
